import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from scipy.optimize import curve_fit

from sklearn.metrics import r2_score

# plt.style.use('./pyDWS/plottingStyle.mplstyle')

import importlib.resources as resources

with resources.path("pyDWS", "plottingStyle.mplstyle") as style_path:
    plt.style.use(style_path)


from pyDWS.plots import plot_layout


def loading_data(dataPath, skipN = 6, save=None):
    """
    loading data

    Parameters
    ----------
    dataPath : string
        Path of the file to load
    skipN : int, optional
        Numver of rows to skip at the beginning of the csv file. The default is 6.

    Returns
    -------
    2 Pandas dataframe of data splitted into a top and bottom part

    """
    # Read the CSV, skipping the first 6 rows (by default)
    dfData_raw = pd.read_csv(dataPath, skiprows=skipN, header=None) 
    
    # Convert everything to float64, coercing errors (like strings) to NaN
    dfData = dfData_raw.apply(pd.to_numeric, errors='coerce')
    
    # Get indices where there's at least one NaN
    nan_rows = dfData[dfData.isna().any(axis=1)].index.tolist()
    
   
    # loading dataframe again, but only second part 
    # dfData_real = pd.read_csv(dataPath, skiprows=nan_rows[0], header=None, dtype=np.float64) 
    
    
    
    # splitting the dataframe into 2 parts
    if nan_rows:
        split_index = nan_rows[0]
        df_top = dfData.iloc[:split_index]
        df_bottom = dfData.iloc[split_index+1 :]
    else:
        df_top = dfData
        df_bottom = pd.DataFrame()  # no NaNs, nothing to split
    
    # saving raw data 
    if save != None:
        if df_bottom.empty:
            raise Exception('icf_raw.txt is empty')
        else:
            df_bottom.to_csv(os.path.join(save, 'icf_raw.txt'), sep='\t', header=['Lag Time [s]', 'Correlation Function'], index=False)
        
    return df_top.reset_index(drop=True), df_bottom.reset_index(drop=True)






def echo_peakFit(echoData, echoPeaks, f, fitModel='Lorentzian', plot_ind=False, savePlot=None, saveEcho=None):
    """
    Fitting the echo peaks using a Lorentzian Model and calculating the area below the peak. 

    Parameters
    ----------
    echoData : pandas dataframe
        Echo data to fit the peaks.
    echoPeaks : numpy array 
        Array which contain the echo peak indices.
    f : float
        frequency of the echo peaks.
    fitModel : String 
        Model to use for the fit. The default is 'Lorentzian'. 'Gaussian' is also predefined.
    plot_ind : boolean or int
        Show plot or not. By default the plot is not shown. 
        When plot is equal an integer the peak at this index is shwon
    save : string
        save the plot or not. The string corresponds to the path where the file is saved or is iqual to None.
        By default it is None. 
    saveEcho : string
        save the Echo data or not. The string corresponds to the path where the file is saved or is iqual to None.
        By default it is None. It saves t the fitting parameters and the area of the fit. 
    Returns
    -------
    df_icf_echo Dataframe which contains t, fit parameters and calculated area.

    """
    # fitting the echo peak and area calculation 

    icf_echo = []
    
    plot = False
    savePlotL = [0, savePlot]
    
    # Fit each peak
    for peak_index in echoPeaks:
        t = peak_index / f
        print('t: ', t)
        mask = np.abs(echoData.iloc[:, 0] - t) < 0.005
        tpeak = echoData[mask].iloc[:, 0]
        gpeak = echoData[mask].iloc[:, 1]
        
        # check wheater plotting or not
        if peak_index == plot_ind:
            plot = True
            savePlotL = [plot_ind, savePlot]
            
            # Plot all echo data
            fig, ax = plt.subplots(figsize=(6.4,4.8))
            ax.plot(echoData.iloc[:,0], echoData.iloc[:,1], 'o', label='Echo measurements')
            ax.plot(tpeak, gpeak, 'o', label='echo peak', color='None', markeredgecolor='red')
            
            plt.legend()
            plot_layout(fig, ax, False, False, 'x', 'y')
            savepath=[str(plot_ind)+'_1', savePlot]
            if savepath[1] != None:
                plt.savefig(os.path.join(savepath[1], 'echo_peak_' + str(savepath[0])+'.png'))
        else:
            plot = False
            

        # Fit using a Lorentzian model by default
        fit_result, gof, r2 = fit_echo_peak(tpeak, gpeak, fitModel, plot, savePlotL)
        
        a = fit_result[0]
        mu = fit_result[1]
        sigma = fit_result[2]

        # calculating area under the fit
        area = a * np.pi * abs(sigma) * 1e4
        
        icf_echo.append([t, a, mu, sigma, area, r2])
        
    df_icf_echo = pd.DataFrame(icf_echo, columns =['t', 'a', 'mu', 'sigma', 'area', 'r2'])
    
    # saving echo data 
    if saveEcho != None:
        df_icf_echo.to_csv(os.path.join(saveEcho, 'icf_echo.txt'), sep='\t', index=False)

    return df_icf_echo





def fit_echo_peak(xData, yData, model, plot, save):
    """
    Fit function to fit the echo peaks. By default it fits the peaks with a Lorentzian. 
    Lorentzian and Gaussian are predefined. It is also possible to fit the data with a custom function.

    Parameters
    ----------
    xData : Dataframe
        One column of a dataframe.
    yData : Dataframe
        One column of a dataframe.
    model : String or custom function
        Model to use for the fit. The default is 'Lorentzian'. 'Gaussian' is also predefined. 
    plot : boolean  
        Default False. 
    save : list or None
        save the plot or not. The first element contains the index of the echo to save and the second element the path where the file is saved.
        By default it is None.
    Returns
    -------
    Fit parameters.

    """
    
    # Initial guess for [amplitude, mean, width]
    initial_guess = [np.max(yData), np.mean(xData), np.std(xData)]
    
    # generate x Data for plotting the fit
    xDataFit = np.linspace(min(xData), max(xData), 100)
    # plot lables
    xlabel = 'x'
    ylabel = 'y'
    
    # Lorentzian Model 
    if model=='Lorentzian':
        # Perform the fit
        print('Lorentzian')
        try:
            popt, pcov = curve_fit(_lorentzian, xData, yData, p0=initial_guess)
            y_pred = _lorentzian(xData, *popt)
            # print('r2: ', r2_score(yData, y_pred))
            r2Score = r2_score(yData, y_pred)
            
        except RuntimeError:
            popt = np.array([0, 0, 0])
            pcov = 0
            r2Score = 0
            print("Error - curve_fit failed")
        
        # plotting data with the fit
        if plot == True:
            fig, ax = plt.subplots(figsize=(6.4,4.8))
            ax.plot(xData, yData, 'o', label='Measurement')
            ax.plot(xDataFit, _lorentzian(xDataFit, popt[0], popt[1], popt[2]), '-', label=model+' Fit')
            
            plt.legend()
            plot_layout(fig, ax, False, False, xlabel, ylabel)
            if save[1] != None:
                plt.savefig(os.path.join(save[1], 'echo_peak_' + str(save[0])+'.png'))
            
            
    # Gaussian Model 
    elif model=='Gaussian':
        # Perform the fit
        print('Gaussian')
        try:
            popt, pcov = curve_fit(_gaussian, xData, yData, p0=initial_guess)
            y_pred = _gaussian(xData, *popt)
            # print('r2: ', r2_score(yData, y_pred))
            r2Score = r2_score(yData, y_pred)
        except RuntimeError:
            popt = np.array([0, 0, 0])
            pcov = 0
            r2Score = 0
            print("Error - curve_fit failed")
        
        # plotting data with the fit
        if plot == True:
            fig, ax = plt.subplots(figsize=(6.4,4.8))
            ax.plot(xData, yData, 'o', label='Measurement')
            ax.plot(xDataFit, _gaussian(xDataFit, popt[0], popt[1], popt[2]), '-', label=model+' Fit')
            
            plt.legend()
            plot_layout(fig, ax, False, False, xlabel, ylabel)
            if save[1] != None:
                plt.savefig(os.path.join(save[1], 'echo_peak_' + str(save[0])+'.png'))

    else:
        raise Exception('Fit could not be performed. Wrong model parameter')
        

    return popt, pcov, r2Score



def generate_echoPeaks():
    """
    Generating echo peaks layout for the echo peak fitting. 
    So the rule is:
    Start with 1–8
    Then generate blocks of 7 numbers with increasing step sizes (2, 3, 4, ..., 11)


    Returns
    -------
    numpy array.

    """
    
    # generating peak distribution 
    all_peak = []
    
    # Subsequent blocks
    current = 1
    for step in range(1, 10):  # Steps from 2 to 11
        for _ in range(7): # Repeat this loop 7 times, but I don’t care about the loop variable
            all_peak.append(current)
            current += step
    all_peak.append(current) # adding one last aditionally 
    
    # converting list to numpy array 
    all_peak_arr = np.array(all_peak, dtype='int32')
    print('Peak distribution: ', all_peak_arr)
    
    

    return all_peak_arr






# =============================================================================
# Private functions
# =============================================================================

# Lorentzian function
def _lorentzian(x, a, mu, sigma):
    """
    Parameters
    ----------
    x : dataframe
        data to fit.
    a : float
        Amplitude.
    mu : float
        center.
    sigma : float
        full width at half maximum.

    Returns
    -------
    TYPE
        Lorentzian function.

    """
    # return a *sigma / (sigma**2 + (x - mu)** 2) # Fully correct Lorentzian 
    return a / (1 + ((x - mu) / sigma)** 2)


# Gaussian function
def _gaussian(x, a, mu, sigma):
    """
    Parameters
    ----------
    x : dataframe
        data to fit.
    a : float
        Amplitude.
    mu : float
        center.
    sigma : float
        full width at half maximum.

    Returns
    -------
    TYPE
        Lorentzian function.

    """
    return a * np.exp(-((x - mu)**2) / (2 * sigma**2))













#

"""
pyDWS
=================

A package for merging and analysing echo and two-cell measurements in DWS.


"""

__version__ = "0.1.0"
__author__ = "M. Helfer, C. Zhang, and F. Scheffold"

# Public imports (main API)
from .data_preprocessing import loading_data, echo_peakFit, fit_echo_peak, generate_echoPeaks
from .utils import generate_EchoLayout, generate_time_array
from .tc_echo_merge import tc_echo_merge
from .icf_fit import icf_fit
from .msd import msd
from .microrheology import microrheology, inertia_correction
from .plots import plot_layout



# Define what gets imported with "from particle_analysis import *"
__all__ = [
    "loading_data",
    "echo_peakFit",
    "fit_echo_peak",
    "generate_echoPeaks",
    "generate_EchoLayout",
    "tc_echo_merge",
    "icf_fit",
    "generate_time_array",
    "msd",
    "microrheology",
    "inertia_correction",
    "plot_layout",
    
]









import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
# from scipy.optimize import minimize
import os
from scipy.optimize import minimize_scalar

# plt.style.use('./pyDWS/plottingStyle.mplstyle')

from pyDWS.plots import plot_layout




# MSD calculation 
def msd(t, g2, lstar, lambda_laser, L, n, radi, range_icf, msdPath=None, savePlots=None):
    """
    Calculating MSD based on:
    D. J. Pine and D. A. Weitz, Diffusing-wave spectroscopy, in
    Dynamic Light Scattering: The Method and Some Applications,
    edited by W. Brown (Oxford University Press, Oxford, 1993)
    pp. 652–720.
    
    C. Zhang, M. Reufer, D. Gaudino, and F. Scheffold, Improved
    diffusing wave spectroscopy based on the automatized determi-
    nation of the optical transport and absorption mean free path,
    Korea-Australia Rheology Journal 29, 241 (2017).
    
    All parameters need to be in meters. 
    
    Parameters
    ----------
    t : numpy array
        time array.
    g2 : numpy array
        ICF -> Intensity correlation function.
    lstar : int
        transport mean free path.
    lambda_laser : float
        laser wavelength λ.
    L : int
        slab thickness.
    n : float
        refractive index.
    radi : float
        radius of the particles in the sample.
    range_icf : list
        time range of the ICF.
    msdPath : string, optional
        Path to save the msd. The default is None.
    savePlots : None or string
        Default None. Otherwise the path where the graphs should be saved. 

    Returns
    -------
    MSD : pandas dataframe
        First column time. Second column calculated mean squared displacement MSD. 
    gcal : numpy array
        Calculated g2 (ICF) function.
    resultL: list
        history of the fit

    """
    
    # setting range of icf
    valid = (g2 > 0) & ~np.isnan(g2)
    tfilt = t[valid]
    g2filt = g2[valid]

    if range_icf is not None:
        low, high = range_icf
        mask = (g2filt > low) & (g2filt < high)
        tfilt = tfilt[mask]
        g2filt = g2filt[mask]

    k0 = 2 * np.pi * n / lambda_laser

    # calculating mid value and start guess
    mi = (min(max(g2filt), 1) - max(min(g2filt), 0)) / 2 + max(min(g2filt), 0)
    r = np.where(g2filt < mi)[0][0]
    ratio = len(g2filt) / r
    idx = int(len(tfilt) / ratio)
    xf, yf = tfilt[idx], g2filt[idx]
    tau0start = -2 * xf * (L**2) / ((lstar**2) * np.log(yf))
    
    # print(idx)
    
    # Initialize empty arrays
    tau = np.zeros(len(tfilt))
    gcal = np.zeros(len(tfilt))
    
    resultL = []
    
    # calculate MSD
    for i in range(idx, -1, -1):
        ti, gi = tfilt[i], g2filt[i]
        
        if gi >= 1 or gi <= 0:
            # print('IN if 1')
            taui = tau[i+1] if i+1 < len(tau) else tau0start
        else:
            fun =  lambda t0: _g2_msd(t0, ti, gi, L, lstar, k0)[0]
            # result = minimize(fun, tau0start, method='Nelder-Mead', options={'xatol': 1e-15, 'maxiter': 10000, 'disp': False})
            result = minimize_scalar(fun, bounds=(1e-4, 1e5), method='bounded', options={'xatol': 1e-16})
            
            # taui = result.x[0]
            taui = result.x
            tau0start = taui
            
            resultL.append(result)
            
        tau[i] = taui
        _, gc = _g2_msd(taui, ti, gi, L, lstar, k0)
        gcal[i] = gc
    
    for i in range(idx + 1, len(tfilt)):
        ti, gi = tfilt[i], g2filt[i]
        if gi >= 1 or gi <= 0:
            # print('IN if 2')
            taui = tau[i - 1]
        else:
            fun =  lambda t0: _g2_msd(t0, ti, gi, L, lstar, k0)[0]
            # result = minimize(fun, tau0start, method='Nelder-Mead', options={'xatol': 1e-15, 'maxiter': 10000, 'disp': False})
            result = minimize_scalar(fun, bounds=(1e-4, 1e5), method='bounded', options={'xatol': 1e-16})
            
            # taui = result.x[0]
            taui = result.x
            tau0start = taui
            
            resultL.append(result)
            
        tau[i] = taui
        _, gc = _g2_msd(taui, ti, gi, L, lstar, k0)
        gcal[i] = gc
    
    
    # transforming t ot msd
    msdC = np.zeros((len(tfilt), 2))
    msdC[:, 0] = tfilt
    msdC[:, 1] = 6 / k0**2 * tfilt / tau
    
    
    
    # Plot g2 and calculated g2
    fig, ax = plt.subplots(figsize=(6.4,4.8))
    ax.plot(tfilt, g2filt, 'ob', label='Measured')
    ax.plot(tfilt, gcal, '-r', label='calculated')
    plt.legend()
    plot_layout(fig, ax, True, False, 'Lag time t (s)', 'ICF')
    if savePlots != None:
        plt.savefig(os.path.join(savePlots, 'recalculated_icf.png'))


    # Plot calculated MSD 
    fig, ax = plt.subplots(figsize=(6.4,4.8))
    ax.plot(msdC[:,0], msdC[:,1] , 'ob')
    plot_layout(fig, ax, True, True, 'Lag time t (s)', r'MSD')
    if savePlots != None:
        plt.savefig(os.path.join(savePlots, 'msd.png'))


    # Saving MSD
    
    pdMSD = pd.DataFrame({'t': msdC[:, 0], 'msd': msdC[:, 1]})
    
    
    if msdPath != None:
        pdMSD.to_csv(os.path.join(msdPath, 'msd.txt'), sep='\t', index=False)

    
    return pdMSD, gcal, resultL




def _g2_msd(tmsd, tr, g2real, L, lstar, k0):
    # defining components from equation
    Ll = L / lstar
    z0l = 2
    # ttau = k0**2 * tmsd
    
    ttau = 6 * tr / tmsd
    beta = 2/3
    
    sqrt_ttau = np.sqrt(ttau)
    
    g1 = (((Ll + 2*beta) / (z0l + beta) * (np.sinh(z0l * sqrt_ttau) + (beta * sqrt_ttau * np.cosh(z0l * sqrt_ttau)))) 
          / (((1 + (beta**2 * ttau)) * np.sinh(Ll * sqrt_ttau)) + (2*beta * sqrt_ttau * np.cosh(Ll * sqrt_ttau))))
    
    
    g2 = g1**2
    
    lerror = ((g2 - g2real) / g2real) ** 2
    
    
    return lerror, g2















import os
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MaxNLocator



# =============================================================================
# Standard Layout of the functions 
# ============================================================================= 
def plot_layout(fig, ax, logX, logY, xlabel, ylabel, title=None):
    """
    Function to set the layout of the plots

    Parameters
    ----------
    fig : figure variable
    ax : axis variable
    logX : True or False
        plot in log or not.
    logY : True or False
        plot in log or not.
    xlabel : string
        label of x axis.
    ylabel : string
        label of y axis .
    title : string
        Ttitle of the plot.

    Returns
    -------
    None.

    """
    
    # set title
    if title != None:
        ax.set_title(title, loc='right', fontsize=15)
    
    
    # log scale of the Plot
    if logY == True:
        ax.set_yscale('log')
        # plt.yscale('log')
    if logX == True:
        ax.set_xscale('log')
        # plt.xscale('log')
    
    

    # On both the X and Y axes, the start and end point of the axis is an integer or power      
    data_xlim = ax.dataLim.x0, ax.dataLim.x1
    data_ylim = ax.dataLim.y0, ax.dataLim.y1
    
    padding_x = 0.02 * (data_xlim[1] - data_xlim[0])
    padding_y = 0.02 * (data_ylim[1] - data_ylim[0])
    
            
    if data_xlim[0] - padding_x < 0:
        xmin = 10**np.floor(np.log10(data_xlim[0]))
        ax.set_xlim(xmin, data_xlim[1] + padding_x)
    else:    
        ax.set_xlim(data_xlim[0] - padding_x, data_xlim[1] + padding_x)
    
    if logY == True:
        if data_ylim[0] - padding_y < 0:
            ymin = 10**np.floor(np.log10(data_ylim[0]))
            ax.set_ylim(ymin, data_ylim[1] + padding_y)
        else:    
            ax.set_ylim(data_ylim[0] - padding_y, data_ylim[1] + padding_y)
    else:
        ax.set_ylim(data_ylim[0] - padding_y, data_ylim[1] + padding_y)
    
        
    # Set major ticks to be at edges
    if logX == False and logY == False:
        ax.xaxis.set_major_locator(MaxNLocator(nbins='auto', steps=[1, 2, 5, 10], prune=None))
        ax.yaxis.set_major_locator(MaxNLocator(nbins='auto', steps=[1, 2, 5, 10], prune=None))
    
    
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    # tight layout = everthing of of the plot must be insede the Figure
    fig.tight_layout()

    
    # saving graph
    # if savePath[1] != None:
    #     plt.savefig(os.path.join(savePath[1], 'echo_peak_' + str(savePath[0])+'.png'), transparent=True)








import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.optimize import minimize
import os
import sys
import json
from pathlib import Path

# plt.style.use('./pyDWS/plottingStyle.mplstyle')

from pyDWS.plots import plot_layout
from pyDWS.utils import generate_time_array


def icf_fit(data, name, alphaRange=np.arange(0, 22.5, 1.5), plateauVal=False, max_iterations=400, iteration_step = 10000, PathPguess= None, icfFitPath=None, savePlots=None):
    """
    Fitting icf with a intercept-adjusted exponential spectrum fit

    Parameters
    ----------
    data : string or pandas Dataframe
        Path of the data to fit or the data as a pandas Dataframe
    name : string
        Name of the Pguess file.
    alphaRange : numpy array, optional
        Alpha range used for the fit. The default is np.arange(0, 22.5, 1.5).
    plateauVal : Bool, optional
        Can be used if data shows a plateau for a better fit. The default is False.
    max_iterations : int, optional
        The minimization must converge under max_iterations. The default is 400.
    iteration_step : int, optional
        How many interations the minimization does in one minimization loop. The default is 10000.
    PathPguess : string, optional
        Paht to save the optimized P values. The default is None.
    icfFitPath : string, optional
        Path to save the fitted icf. The default is None.
    savePlots : None or string
        Default None. Otherwise the path where the graphs should be saved. 

    Returns
    -------
    TYPE
        dictionary.
        icfFitData = {
            't': tN, (lag times)
            'icf': _ICFfit(tN, g2, alp, P_current), (fitted icf)
            'intercept': intercept, (to normalize to 1)
            'alpha': alp, (alpha values)
            'P': P_current, (optimized P values)
            'fun': functionVal, (function values of the minimization)
            'NIter': numberInter, (number of minimization loops)
            'MinIter': total_iterations (number of iterations in the last minimization loop)
        }

    """
    
    # Load data
    if isinstance(data, pd.DataFrame):
        t = data.iloc[:,0].values
        g2 = data.iloc[:,1].values
    elif isinstance(data, (str, Path)):
        ICF = pd.read_csv(data, sep='\t')
        t = ICF.iloc[:,0].values
        g2 = ICF.iloc[:,1].values
    else:
        raise TypeError(f"Expected a pandas DataFrame or path (str/Path), got {type(data).__name__}")
    
    
    # Alpha array
    alp = 1e-7 * np.exp(alphaRange)

    if plateauVal == True:
        alp[-1] = np.inf
        print('alpha:', alp)
    
    
    
    # Load previous guess or create new
    try:
        guess_path = os.path.join(PathPguess, f'ICFfitguess_{name}.npy')
    except TypeError:
        print("Error - Invalid path")
        sys.exit()
    
    if os.path.exists(guess_path):
        Pguess = np.load(guess_path)
    else:
        Pguess = np.random.rand(len(alp))
        Pguess = Pguess / np.sum(Pguess)



    # minimization parameters
    P_current = Pguess.copy()
    converged = False
    total_iterations = 10000
    history = [P_current.copy()]
    functionVal = []

    numberInter = 0

    # objectiv function used for minimization 
    def _objective(P):
       return _ICFfitMin(t, g2, alp, P)


    # minimize objectiv function 
    while not converged or (total_iterations > max_iterations):
    
        numberInter +=1
        
        result = minimize(_objective, P_current, method='Nelder-Mead', options={'maxiter': iteration_step, 'disp': False})
    
        P_current = result.x
        total_iterations = result.nit
        history.append(P_current.copy())
    
        # P_current = P_new
        converged = result.success
        functionVal.append(result.fun)
        
        
    
    # if not converged:
    #     print(f'Max iterations reached ({total_iterations}). Final ΔP={delta:.2e}')
    if PathPguess != None:
        np.save(guess_path, P_current)
        
    # Calculating Intercpt
    intercept = np.sum(P_current**2)


    # generate time array as for multitau measurement
    tN = generate_time_array()


    # Plot evolution of function values
    fig, ax = plt.subplots(figsize=(6.4,4.8))
    ax.plot(np.arange(1, numberInter + 1, 1), functionVal , 'o')
    plot_layout(fig, ax, False, False, 'Iteration', 'Function value')
    
    
    
    # Plot fit
    fig, ax = plt.subplots(figsize=(6.4,4.8))
    ax.plot(t, g2 , 'o', label='merged')
    ax.plot(tN, _ICFfit(tN, g2, alp, P_current), '-', label='icf Fit')
    plt.legend()
    plot_layout(fig, ax, True, False, 'Lag time t (s)', r'ICF')
    if savePlots != None:
        plt.savefig(os.path.join(savePlots, 'icf_fit.png'))
    
    
    
    # Saving Fit
    icfFitData = {
        't': tN.tolist(), 
        'icf': _ICFfit(tN, g2, alp, P_current).tolist(),
        'intercept': intercept,
        'alpha': alp.tolist(), 
        'P': P_current.tolist(), 
        'fun': functionVal, 
        'NIter': numberInter,
        'MinIter': total_iterations
    }
    
    if icfFitPath != None:
        with open(os.path.join(icfFitPath, "icf_fit.txt"), "w") as f:
            json.dump(icfFitData, f, indent=4)
    
    

    
    
    return icfFitData
    



# Fit function to minimize 
def _ICFfitMin(t, g2, alp, P, ifweight=False):
    
    ICFfit = np.zeros_like(t)
    f2 = np.zeros_like(t)
    
    for i in range(len(t)):
        ICFfit[i] = np.sum(P**2 * np.exp(-t[i]/alp)**2)
        f2[i] = np.sum(ICFfit[i] / alp**2)
    
    residual = ICFfit - g2
    sigma = np.std(residual) + np.finfo(float).eps
    
    if ifweight:
        weights = 1 / (sigma**1)
        Er = np.mean(weights * (ICFfit - g2)**2)
    else:
        Er = np.mean((ICFfit - g2)**2)
    
    return Er

# ICF Fit
def _ICFfit(t, g2, alp, P):
    
    ICFfit = np.zeros_like(t)
    f2 = np.zeros_like(t)
    
    for i in range(len(t)):
        ICFfit[i] = np.sum(P**2 * np.exp(-t[i]/alp)**2)
        f2[i] = np.sum(ICFfit[i] / alp**2)
    
    
    return ICFfit













